# coding: utf-8

import argparse
import os
import sys
from sys import platform

import logging

sys.tracebacklimit = 0

# pylint: disable=wrong-import-position
from tionix_vdi_client import MODULE_VERSION  # nopep8
from tionix_vdi_client.run_self_diagnostics import run_diagnostics  # nopep8
from tionix_vdi_client.settings import (  # nopep8
    ConfigSettings, SessionSettings)
from tionix_vdi_client.ui_tk import start_gui  # nopep8
from tionix_vdi_client.vdi_connect import vdi_connect  # nopep8

logger = logging.getLogger(__name__)


def main():
    exit_code = 0
    if platform == "win32":
        start_gui(vdi_connect)
    else:
        args = create_parser().parse_args()
        if args.diagnostic:
            debug = args.debug
            if debug:
                del sys.tracebacklimit
            run_diagnostics(debug=debug)
        elif args.no_ui:
            exit_code = connect_without_ui(args)
        else:
            start_gui(vdi_connect)
    return exit_code


def connect_without_ui(args):
    session_settings = SessionSettings()
    config_settings = ConfigSettings()

    session_settings.read()
    config_settings.read(reread=True)

    username = args.username or session_settings.user
    password = args.password or session_settings.password
    project = args.project or session_settings.project

    domain = (args.domain or
              session_settings.domain_name or
              config_settings.domain_name)

    server_uri = (args.server_uri or
                  session_settings.cloud or
                  config_settings.cloud)

    connect = (args.connect or
               session_settings.connect or
               config_settings.connect)

    store_session = args.store_session

    conn_credentials = {
        'user': username,
        'password': password,
        'project': project,
        'domain_name': domain,
    }

    session_settings.update(conn_credentials)

    if store_session:
        session_settings.save()

    config_settings.cloud = server_uri
    config_settings.connect = connect

    try:
        vdi_connect(conn_credentials, config_settings)
    except Exception as err:
        logger.error(err)
        return getattr(os, 'EX_SOFTWARE', 1)

    return getattr(os, 'EX_SOFTWARE', 0)


def create_parser():
    description = u'TIONIX.VDIclient utility'
    parser = argparse.ArgumentParser(description=description,
                                     conflict_handler='resolve')
    parser.add_argument(
        '--version', action='version', version=MODULE_VERSION)
    parser.add_argument(
        '--username',
        required=False,
        type=str,
        help='User name for connection',
    )
    parser.add_argument(
        '--password',
        required=False,
        type=str,
        help='Password for connection',
    )
    parser.add_argument(
        '--project',
        required=False,
        type=str,
        help='Project for vm instance retrieval',
    )
    parser.add_argument(
        '--domain',
        required=False,
        type=str,
        help='Project domain name',
    )
    parser.add_argument(
        '--server-uri',
        required=False,
        type=str,
        help='Uri for vm instance retrieval',
    )
    parser.add_argument(
        '--connect',
        required=False,
        type=str,
        help='RDP-client connection command template. '
             'Example: xfreerdp /u:{user} /p:{password}  /v:{ip}',
    )
    parser.add_argument(
        '--no-ui',
        action='store_true',
        required=False,
        help='Provide this flag if there is no need to launch GUI',
    )
    parser.add_argument(
        '--store-session',
        action='store_true',
        required=False,
        help='Provide this flag if there is a need to save user, password, '
             'project, domain to session file',
    )
    parser.add_argument(
        '--diagnostic',
        action='store_true',
        help='Run self-diagnostic'
    )
    parser.add_argument(
        '--debug',
        dest='debug',
        action='store_true',
        help='Activate the debug mode for self-diagnostic'
    )
    return parser

if __name__ == '__main__':
    sys.exit(main())
