# coding: utf-8

import functools
import logging
import sys

from collections import namedtuple

from tionix_vdi_client.i18n import ugettext as _

try:
    import tkMessageBox as messagebox
except ImportError:
    messagebox = None


logger = logging.getLogger(__name__)


class VDIException(Exception):
    pass


class VDIServerResponseError(VDIException):
    pass


class VDIServerBadCredentialsError(VDIServerResponseError):
    pass


class VDIServerBadConnectionParamsError(VDIServerResponseError):
    pass


class ApplicationLogicException(VDIException):
    pass


class MissingParameterException(VDIException):
    pass


class ApplicationConnectionError(VDIException):
    pass

AlertMessage = namedtuple(
    'AlertMessage',
    ['exception', 'title']
)


ALERT_MESSAGES = (
    AlertMessage(
        exception=ApplicationLogicException,
        title=_(u'Connection error'),
    ),
    AlertMessage(
        exception=MissingParameterException,
        title=_(u'Connection error'),
    ),
    AlertMessage(
        exception=ApplicationConnectionError,
        title=_(u'Connection error'),
    ),
    AlertMessage(
        exception=ValueError,
        title=_(u'Invalid input error'),
    ),
    AlertMessage(
        exception=VDIServerBadCredentialsError,
        title=_(u'Invalid credentials'),
    ),
    AlertMessage(
        exception=VDIServerBadConnectionParamsError,
        title=_(u'Invalid connection parameters'),
    ),
    AlertMessage(
        exception=VDIServerResponseError,
        title=_(u'Server response error'),
    ),
)


def process_exception(err, fall=True):
    """Exception handler

    Log error and show alert message
    """
    exc = type(err)
    message = err.message or str(err)

    for alert_message in ALERT_MESSAGES:
        if issubclass(exc, alert_message.exception):
            title = alert_message.title
            break
    else:
        title = _(u'Error')

    logger.error(message)
    if messagebox:
        try:
            messagebox.showerror(title, message)
        except Exception:
            if sys.platform == 'win32':
                from tionix_vdi_client.winapi import message_box
                message_box(message, title)

    if fall:
        sys.exit(1)


def catch_exceptions(fall=True):
    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                f(*args, **kwargs)
            except Exception as err:
                process_exception(err, fall)
        return wrapped
    return wrapper
