# coding: utf-8

import base64
import getpass

import locale
import logging
import logging.config
import os
import shutil
import sys
import tempfile

import ConfigParser as configparser

from tionix_vdi_client.helpers import Localization, get_key_by_value

APP_DIR_NAME = '.tionix-vdi-client'
APP_DIR_PATH = os.path.join(os.path.expanduser('~'), APP_DIR_NAME)

if getattr(sys, 'frozen', False):
    # The application is frozen
    BASE_DIR = os.path.dirname(sys.executable)
else:
    BASE_DIR = os.path.dirname(__file__)

BASE_DIR = BASE_DIR.decode(sys.getfilesystemencoding())

RESOURCES_DIR = os.path.join(BASE_DIR, 'resources')

_settings = None  # Settings cache if reread settings is not required

LOG_FILE_NAME = 'tionix-vdi-client'
LOG_FILE_EXT = 'log'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'default': {
            'format': '%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
                      '%(name)s [-] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        'extended': {
            'format': '%(asctime)s - %(levelname)s - File: %(filename)s - '
                      '%(funcName)s() - Line: %(lineno)d - %(message)s'
        }
    },
    'handlers': {
        'null': {
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'default',
        },
    },
    'loggers': {
        'tionix_vdi_client': {
            'level': 'INFO',
            'handlers': ['console']
        },
    }
}

logging.config.dictConfig(LOGGING)

logger = logging.getLogger(__name__)


def get_default_connect_param():
    if sys.platform == "win32":
        conn_cmd = 'mstsc {rdp_config}'
    else:
        conn_cmd = 'xfreerdp /u:{user} /p:{password}  /v:{ip}'
    return conn_cmd


def get_default_locale():
    code, _encoding = locale.getdefaultlocale()
    return (get_key_by_value(Localization.locales_dict, code)
            if code else Localization.EN)


def validate_log_level(v, def_v):
    return def_v if isinstance(logging.getLevelName(v), basestring) else v


def validate_bool(val, default_value=False):
    if val in (True, 'True', 'yes', 'Yes' 'y', '1', 1):
        value = True
    elif val in (False, 'False', 'no', 'No', 'n', '0', 0):
        value = False
    else:
        value = default_value
    return value


def validate_str(val, def_v=''):
    val = val.strip()
    if not val:
        val = def_v
    return val


def validate_log_file_location(val, def_value):
    result = def_value
    if val:
        if isinstance(val, str):
            val = val.decode('utf-8')
        val = os.path.expandvars(val)
        if not os.path.exists(val):
            os.makedirs(val)

        if os.path.isdir(val):
            result = val
    return result


def validate_language(val, def_value):
    result = def_value
    available_languages = Localization.locales_dict.keys()
    if isinstance(val, basestring):
        val = val.lower().strip()
        if val in available_languages:
            result = val
        else:
            logger.warning(u'Language "{0}" is not supported. Available '
                           u'languages are: "{1}". Your system language '
                           u'is "{2}", so it is set as the default '
                           u'language.'.format(val,
                                               u', '.join(available_languages),
                                               def_value))
    return result


def get_log_file_path(dir_path=None):
    dir_path = os.path.abspath(dir_path or APP_DIR_PATH)

    current_user = getpass.getuser()

    available_names = (
        '{0}.{1}'.format(LOG_FILE_NAME, LOG_FILE_EXT),
        '{0}-{1}.{2}'.format(LOG_FILE_NAME, current_user, LOG_FILE_EXT),
    )

    for name in available_names:
        if os.access(dir_path, os.W_OK):
            path = os.path.join(dir_path, name)
            if not os.path.exists(path) or os.access(path, os.W_OK):
                break
    else:
        _id, path = tempfile.mkstemp(
            prefix=LOG_FILE_NAME + '-', suffix='.' + LOG_FILE_EXT)
    return path


def update_nested_dict(initial_dict, update_dict):
    for key, val in update_dict.items():
        if isinstance(val, dict):
            initial_dict[key] = update_nested_dict(
                initial_dict.get(key, {}), val)
        else:
            initial_dict[key] = val
    return initial_dict


class Settings(dict):
    DOMAIN_NAME = 'domain_name'
    CLOUD = 'cloud'
    CONNECT = 'connect'
    LANGUAGE = 'language'
    required_attributes = ()

    _attributes = [DOMAIN_NAME, CLOUD, CONNECT, LANGUAGE]

    def __getattr__(self, name):
        if name in self._attributes:
            return self.get(name, '')
        else:
            msg = u"'{0}' object has no attribute '{1}'".format(
                type(self).__name__, name)
            raise AttributeError(msg)

    def __setattr__(self, item, value):
        if item in self._attributes:
            self[item] = value
        else:
            msg = u"'{0}' object has no attribute '{1}'".format(
                type(self).__name__, item)
            raise AttributeError(msg)


class SessionSettings(Settings):
    USER = 'user'
    PASSWORD = 'password'
    PROJECT = 'project'
    required_attributes = (USER, PASSWORD, PROJECT, Settings.DOMAIN_NAME)

    _attributes = Settings._attributes + [USER, PASSWORD, PROJECT]

    _encode_keys = (USER, PASSWORD, PROJECT, Settings.DOMAIN_NAME)
    _session_section_name = 'DEFAULT'

    @property
    def _session_file_name(self):
        return os.path.join(APP_DIR_PATH, 'tnx_client_session')

    def read(self):
        if os.path.isfile(self._session_file_name):
            config = configparser.ConfigParser()
            try:
                with open(self._session_file_name, 'r') as session_fp:
                    config.readfp(session_fp)
                    for key, value in config.items(self._session_section_name):
                        if key in self._encode_keys:
                            val = base64.decodestring(value)
                        else:
                            val = value
                        setattr(self, key, val)
            except IOError as err:
                logger.warning(u'Error while reading session '
                               u'file: {0}'.format(err))

    def save(self):
        config = configparser.ConfigParser()
        for key, value in self.items():
            if key in self._encode_keys:
                val = base64.encodestring(value)
            else:
                val = value
            config.set(self._session_section_name, key, val)
        try:
            with open(self._session_file_name, 'w') as session_fh:
                config.write(session_fh)
        except IOError as err:
            logger.warning(u'Error while saving session to '
                           u'file: {0}'.format(err))

    def clear(self):
        for attr in self._attributes:
            setattr(self, attr, '')


class ConfigSettings(Settings):
    LOG_LEVEL = 'log_level'
    STORE_SESSION = 'store_session'
    LOG_FILE_LOCATION = 'log_file_location'
    required_attributes = (Settings.CLOUD, Settings.CONNECT)

    _attributes = Settings._attributes + [LOG_LEVEL, STORE_SESSION,
                                          LOG_FILE_LOCATION]

    _config_section_name = 'DEFAULT'
    _conf_file = 'client.conf'
    _sample_file = 'client.conf.sample'

    ordered_file_variables = (
        Settings.CLOUD,
        Settings.DOMAIN_NAME,
        LOG_LEVEL,
        Settings.CONNECT,
        LOG_FILE_LOCATION,
        STORE_SESSION,
        Settings.LANGUAGE,
    )

    default = {
        Settings.CLOUD: '',
        Settings.CONNECT: get_default_connect_param(),
        Settings.DOMAIN_NAME: 'default',
        LOG_LEVEL: 'INFO',
        STORE_SESSION: True,
        LOG_FILE_LOCATION: '',
        Settings.LANGUAGE: get_default_locale(),
    }

    _validators = {
        Settings.CONNECT: validate_str,
        LOG_LEVEL: validate_log_level,
        STORE_SESSION: validate_bool,
        LOG_FILE_LOCATION: validate_log_file_location,
        Settings.LANGUAGE: validate_language,
    }

    @property
    def _sample_config_name(self):
        return os.path.join(BASE_DIR, self._sample_file)

    @property
    def _default_config_name(self):
        return os.path.join(APP_DIR_PATH, self._conf_file)

    def __init__(self):
        super(ConfigSettings, self).__init__()
        self.create_configuration_file()

    def create_configuration_file(self):
        if not os.path.isfile(self._default_config_name):
            try:
                shutil.copy(self._sample_config_name,
                            self._default_config_name)
                self.save()
            except IOError:
                logger.info(u"Couldn't create config file, using defaults!")

    @classmethod
    def validator(cls, setting_param):
        return cls._validators.get(setting_param, lambda x, def_x: x)

    def _read_config(self):
        config = SafeParser()
        try:
            files_parsed = config.read(self._default_config_name)
        except (configparser.MissingSectionHeaderError,
                configparser.ParsingError):
            files_parsed = None
        if not files_parsed:
            logger.warning(u'Error while reading config file. '
                           u'Settings set to default.')
            result = self.default
        else:
            result = {}
            for setting_param in self.ordered_file_variables:
                validator = self.validator(setting_param)
                default = self.default[setting_param]
                result[setting_param] = validator(
                    config.safe_get(self._config_section_name, setting_param,
                                    default=default), default)
        return result

    def read(self, reread=False):
        """
        Read the settings file
        """
        if not reread and self:
            return self
        self.update(self.default)
        self.update(self._read_config())
        logging.config.dictConfig(get_logging_config(
            level=self[self.LOG_LEVEL],
            log_file_dir_path=self[self.LOG_FILE_LOCATION]))

    def save(self):
        """
        Save settings to file
        """
        config = configparser.ConfigParser()
        if self.items():
            for key in self.ordered_file_variables:
                value = getattr(self, key)
                if isinstance(value, unicode):
                    # config doesn't write to file unicode values
                    value = value.encode('utf-8')
                config.set(self._config_section_name, key, value)
        try:
            with open(self._default_config_name, 'w') as config_fh:
                config.write(config_fh)
        except IOError as err:
            logger.warning(u'Error while saving settings to '
                           u'file: {0}'.format(err))


class SafeParser(configparser.ConfigParser):
    def safe_get(self, *args, **kwargs):
        val = kwargs.pop('default', None)
        try:
            val = self.get(*args, **kwargs)
        except configparser.NoOptionError:
            pass
        return val


def get_logging_config(level=None, log_file_dir_path=None):
    level = level or ConfigSettings.default[ConfigSettings.LOG_LEVEL]
    log_file_dir_path = log_file_dir_path or ConfigSettings.default[
        ConfigSettings.LOG_FILE_LOCATION]

    previous_log_path = LOGGING.get('handlers', {}).get(
        'file', {}).get('filename')

    if (previous_log_path and os.path.abspath(
            os.path.dirname(previous_log_path)) == os.path.abspath(
                log_file_dir_path)):
        log_file_path = previous_log_path
    else:
        log_file_path = get_log_file_path(log_file_dir_path)

    if previous_log_path:
        if previous_log_path != log_file_path:
            logger.info(u'Log file location changed from '
                        u'"{0}" to "{1}"'.format(previous_log_path,
                                                 log_file_path))
    else:
        logger.info(u'Log file path: {0}'.format(log_file_path))

    custom_config = {
        'handlers': {
            'null': {
                'level': level,
            },
            'console': {
                'level': level,
            },
            'file': {
                'class': 'logging.FileHandler',
                'filename': log_file_path,
                'formatter': 'default',
                'encoding': 'utf-8',
            }
        },
        'loggers': {
            'tionix_vdi_client': {
                'level': level,
                'handlers': ['console', 'file']
            },
        }
    }
    update_nested_dict(LOGGING, custom_config)
    return LOGGING


# Create directory for application files
if not os.path.exists(APP_DIR_PATH):
    os.mkdir(APP_DIR_PATH)

ConfigSettings().read(reread=True)
