# coding: utf-8
from __future__ import unicode_literals

import logging
import os
import datetime

from sys import platform as _platform
from threading import Timer

import Tkinter as tk
import ttk
import tkMessageBox as messagebox

from tionix_vdi_client import MODULE_VERSION, WIN_OS_BITNESS
from tionix_vdi_client import elements
from tionix_vdi_client.exceptions import catch_exceptions
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.resources.icon import favicon
from tionix_vdi_client.helpers import Localization, get_key_by_value, Colors
from tionix_vdi_client.settings import (
    ConfigSettings, RESOURCES_DIR, SessionSettings
)


logger = logging.getLogger(__name__)

MAX_TEXT_LENGTH = 30
COPYRIGHTS_TEXT = '© Copyright 2015-%s' % datetime.date.today().year
COMPANY = "TIONIX"
TIONIX_WEB_LINK = 'http://tionix.ru'
SUPPORT_EMAIL = 'support@tionix.ru'
HELP_LINK = (
    'http://docs.tionix.ru/tionix_modules/tionix-vdi-client-docs/index.html')


def character_limit(entry):
    entry_text = entry.get()
    if entry_text:
        entry.set(entry_text[:MAX_TEXT_LENGTH])


@catch_exceptions(fall=True)
def start_gui(vdi_connect_callback):
    """Starting point when module is the main routine."""
    root = tk.Tk()
    if _platform == "win32":
        root.tk.call('encoding', 'system', 'cp1252')
    TionixVDIClient(root, vdi_connect_callback)
    root.mainloop()


class TionixVDIClient(object):
    def __init__(self, master=None, vdi_connect_callback=None):
        self.config_settings = ConfigSettings()
        self.session_settings = SessionSettings()

        self.config_settings.read(reread=True)

        self.root = master
        self.vdi_connect = vdi_connect_callback

        if self.config_settings.store_session:
            self.session_settings.read()

        self.timer_to_enable = None
        self.raw_passwd = False
        self.root.configure(takefocus='0')
        self.root.title(_('TIONIX.VDIclient'))
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)

        self._configure_grid()
        self._set_geom()
        self._bind_events()
        self._configure_elements()
        self._set_icon()

    @staticmethod
    def _get_element_row(element_type):
        _elements_row = {
            'help': 1,
            'user': 2,
            'password': 3,
            'group': 4,
            'save_parameters': 5,
            'connect': 6,
            'settings': 6,
            'version': 7,
            'copyrights': 8,
        }
        return _elements_row.get(element_type, 0)

    @staticmethod
    def _get_element_column(element_type):
        _elements_columns = dict(label=1, entry=2, help=2)
        return _elements_columns.get(element_type, 0)

    def _configure_grid(self):
        self.root.grid_rowconfigure(self._get_element_row('user'), pad=25)
        self.root.grid_rowconfigure(self._get_element_row('password'), pad=5)
        self.root.grid_rowconfigure(self._get_element_row('group'), pad=25)
        self.root.grid_rowconfigure(self._get_element_row('connect'), pad=5)
        self.root.grid_rowconfigure(self._get_element_row('version'), pad=5)

        self.root.grid_columnconfigure(self._get_element_column('label'),
                                       minsize=140)

    def _calculate_geom(self):
        ws = self.root.winfo_screenwidth()
        hs = self.root.winfo_screenheight()
        w = self.root.winfo_reqwidth()
        h = self.root.winfo_reqheight()
        x = (ws / 2) - (w / 2)
        y = (hs / 2) - (h / 2)
        return '%dx%d+%d+%d' % (w, h, x, y)

    def _set_geom(self):
        geom = self._calculate_geom()
        # self.root.geometry(geom)
        self.root.geometry('+%s' % geom.split('+', 1)[1])

    def _bind_events(self):
        self.root.bind('<Return>', self.on_return_event)
        self.root.bind('<KP_Enter>', self.on_return_event)
        self.root.bind('<Escape>', self.on_esc_event)
        self.root.protocol("WM_DELETE_WINDOW", self.destroy_window)

    def _configure_elements(self):
        # pylint: disable=too-many-statements
        self.HelpLabel = elements.HelpLabel(
            self.root,
            row=self._get_element_row('help'),
            col=self._get_element_column('help'),
            link=HELP_LINK,
        )

        self.UserLabel = tk.Label(self.root)
        self.UserLabel.configure(text=_('Username'))
        self.UserLabel.grid(row=self._get_element_row('user'),
                            column=self._get_element_column('label'),
                            sticky=tk.W, padx=10)

        self.PasswordLabel = tk.Label(self.root)
        self.PasswordLabel.configure(text=_('Password'))
        self.PasswordLabel.grid(row=self._get_element_row('password'),
                                column=self._get_element_column('label'),
                                sticky=tk.W, padx=10)

        self.GroupLabel = tk.Label(self.root)
        self.GroupLabel.configure(text=_('Project'))
        self.GroupLabel.grid(row=self._get_element_row('group'),
                             column=self._get_element_column('label'),
                             sticky=tk.W, padx=10)

        self.UserEntry = elements.CustomEntry(self.root)
        self.UserEntry.configure(background='white')
        self.UserEntry.configure(font='TkFixedFont')
        self.UserEntry.configure(takefocus='1')
        self.UserEntry.grid(row=self._get_element_row('user'),
                            column=self._get_element_column('entry'),
                            sticky=tk.E, padx=10)
        self.UserEntry.focus_set()
        self.UserEntry.insert(0, self.session_settings.user)

        self.PasswordEntry = elements.CustomEntry(self.root)
        self.PasswordEntry.configure(background='white')
        self.PasswordEntry.configure(font='TkFixedFont')
        self.PasswordEntry.configure(takefocus='1')
        self.PasswordEntry.configure(show='*')
        self.PasswordEntry.grid(row=self._get_element_row('password'),
                                column=self._get_element_column('entry'),
                                sticky=tk.E, padx=10)
        self.PasswordEntry.insert(0, self.session_settings.password)

        self.GroupEntry = elements.CustomEntry(self.root)
        self.GroupEntry.configure(background='white')
        self.GroupEntry.configure(font='TkFixedFont')
        self.GroupEntry.configure(takefocus='1')
        self.GroupEntry.grid(row=self._get_element_row('group'),
                             column=self._get_element_column('entry'),
                             sticky=tk.E, padx=10)
        self.GroupEntry.insert(0, self.session_settings.project)

        self.SessionCheckBox = elements.SaveParametersCheckBox(
            self.root,
            row=self._get_element_row('save_parameters'),
            column=self._get_element_column('entry'),
            default_value=self.config_settings.store_session,
        )

        self.SettingsButton = tk.Button(self.root)
        self.SettingsButton.configure(activebackground=Colors.GREY)
        self.SettingsButton.configure(command=self.on_settings_clicked)
        self.SettingsButton.configure(takefocus='1')
        self.SettingsButton.configure(text=_('Settings'))
        self.SettingsButton.grid(row=self._get_element_row('settings'),
                                 column=self._get_element_column('label'),
                                 sticky=tk.W + tk.N, padx=10, pady=10)

        self.ConnectButton = tk.Button(self.root)
        self.ConnectButton.configure(activebackground=Colors.GREY)
        self.ConnectButton.configure(command=self.on_connect_clicked)
        self.ConnectButton.configure(takefocus='1')
        self.ConnectButton.configure(text=_('Connect'))
        self.ConnectButton.grid(row=self._get_element_row('connect'),
                                column=self._get_element_column('entry'),
                                sticky=tk.E + tk.N, padx=10, pady=10)

        self.PasswordButton = tk.Button(self.root)
        self.PasswordButton.configure(activebackground=Colors.GREY)
        self.PasswordButton.configure(command=self.on_show_password_clicked)
        self.PasswordButton.grid(
            row=self._get_element_row('password'),
            column=self._get_element_column('entry'),
            sticky=tk.E, padx=10, pady=10)
        self.eye_opened = tk.PhotoImage(
            file=os.path.join(RESOURCES_DIR, "opened.gif"))
        self.eye_crossed = tk.PhotoImage(
            file=os.path.join(RESOURCES_DIR, "crossed.gif"))

        self.VersionLabel = tk.Label(self.root)
        self.VersionLabel.configure(text=(_('Version {0}{1}')).format(
            MODULE_VERSION, WIN_OS_BITNESS))
        self.VersionLabel.grid(row=self._get_element_row('version'),
                               column=self._get_element_column('label'),
                               columnspan=2)

        self.CopyrightsFrame = tk.Frame(self.root)
        self.CopyrightsFrame.grid(row=self._get_element_row('copyrights'),
                                  column=self._get_element_column('label'),
                                  columnspan=2)
        self.CopyrightLabel = tk.Label(self.root)
        self.CopyrightLabel.configure(text=COPYRIGHTS_TEXT + ',',
                                      font=(None, 8))
        self.CopyrightLabel.grid(row=1, column=1, in_=self.CopyrightsFrame)

        self.WebsiteLabel = elements.LinkLabel(self.root)
        self.WebsiteLabel.configure(text=COMPANY + ',', font=(None, 8))
        self.WebsiteLabel.grid(row=1, column=2, in_=self.CopyrightsFrame)
        self.WebsiteLabel.bind_link(TIONIX_WEB_LINK)

        self.SupportLabel = elements.LinkLabel(self.root)
        self.SupportLabel.configure(text=SUPPORT_EMAIL, font=(None, 8))
        self.SupportLabel.grid(row=1, column=3, in_=self.CopyrightsFrame)
        self.SupportLabel.bind_link('mailto:' + SUPPORT_EMAIL)

        self.toogle_image()

    def _set_icon(self):
        img = tk.PhotoImage(data=favicon)
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

    def on_return_event(self, event):
        if not self.timer_to_enable or self.timer_to_enable.finished.is_set():
            self.on_connect_clicked()
        else:
            pass

    def on_esc_event(self, event):
        self.destroy_window()

    def destroy_window(self):
        if messagebox.askyesno(_('Exit'), _('Are you sure you want to quit?')):
            self.root.destroy()
            self.root = None

    def enable_connect_button(self):
        if self.root:
            self.ConnectButton.configure(state='normal')

    def get_validated_auth_params(self):
        user = self.UserEntry.get().encode('utf-8')
        password = self.PasswordEntry.get().encode('utf-8')
        group = self.GroupEntry.get().encode('utf-8')
        if all((not user, not password, not group)):
            message = _('Please make sure all required fields are filled in.')
            raise ValueError(message)
        message = _(
            'Invalid value provided in field: "%s".\n'
            'Please enter a valid value.'
        )
        if not user:
            message = message % self.UserLabel.cget('text')
            raise ValueError(message)
        if not password:
            message = message % self.PasswordLabel.cget('text')
            raise ValueError(message)
        if not group:
            message = message % self.GroupLabel.cget('text')
            raise ValueError(message)
        return user, password, group

    @catch_exceptions(fall=False)
    def on_connect_clicked(self):
        user, password, group = self.get_validated_auth_params()
        self.config_settings.read(reread=True)
        self.config_settings.store_session = self.SessionCheckBox.value.get()
        self.config_settings.save()

        conn_credentials = {
            'user': user,
            'password': password,
            'project': group,
            'domain_name': self.config_settings.domain_name,
        }

        if self.config_settings.store_session:
            self.session_settings.update(conn_credentials)
        else:
            self.session_settings.clear()
        self.session_settings.save()

        self.timer_to_enable = Timer(5, self.enable_connect_button)
        self.ConnectButton.configure(state='disabled')
        self.timer_to_enable.start()
        self.vdi_connect(conn_credentials, self.config_settings)

    def toogle_image(self):
        if self.raw_passwd:
            self.PasswordButton.config(image=self.eye_crossed,
                                       width="12", height="12")
        else:
            self.PasswordButton.config(image=self.eye_opened,
                                       width="12", height="12")

    def on_show_password_clicked(self):
        if self.raw_passwd:
            self.PasswordEntry.configure(show='*')
            self.raw_passwd = False
        else:
            self.PasswordEntry.configure(show='')
            self.raw_passwd = True

        self.toogle_image()

    @staticmethod
    def wait_modal_close(root_window, modal_window):
        def handler():
            modal_window.focus_force()

        if _platform == "win32":
            root_window.wait_window(modal_window)
        else:
            WM_DELETE_WINDOW = 'WM_DELETE_WINDOW'
            WM_TAKE_FOCUS = 'WM_TAKE_FOCUS'
            old_close_handler = root_window.protocol(WM_DELETE_WINDOW)
            old_focus_handler = root_window.protocol(WM_TAKE_FOCUS)

            root_window.protocol(WM_DELETE_WINDOW, handler)
            root_window.protocol(WM_TAKE_FOCUS, handler)

            root_window.wait_window(modal_window)

            root_window.protocol(WM_DELETE_WINDOW, old_close_handler)
            root_window.protocol(WM_TAKE_FOCUS, old_focus_handler)

    @catch_exceptions(fall=False)
    def on_settings_clicked(self):
        settings_root = tk.Toplevel()
        settings_root.withdraw()
        settings_root.update_idletasks()
        settings_root.transient(self.root)
        parent_pos = [
            self.root.winfo_x(),
            self.root.winfo_y(),
            self.root.winfo_width(),
            self.root.winfo_height(),
        ]
        SettingsWindow(settings_root, self.config_settings, parent_pos)
        settings_root.grab_set()
        self.wait_modal_close(self.root, settings_root)


class WidgetLayout(object):
    default_padx = 10
    default_pady = 10

    def __init__(self, row=0, column=0, sticky=None, padx=None, pady=None):
        self.row = row
        self.column = column
        self.sticky = sticky
        self.padx = padx or self.default_padx
        self.pady = pady or self.default_pady


class SettingsWindow(object):
    def __init__(self, master, config_settings, parent_pos):
        self.config_settings = config_settings
        self.config_settings.read(reread=True)
        self.parent_pos = parent_pos
        self.root = master
        self.entry_widgets = {}
        self._configure_widgets()
        self._configure_window()

    def _configure_window(self):
        self.root.title(_('Settings'))
        self.root.resizable(width=tk.FALSE, height=tk.FALSE)

        img = tk.PhotoImage(data=favicon)
        self.root.tk.call('wm', 'iconphoto', self.root._w, img)

        self._set_geom()
        self.root.deiconify()

    def _calculate_geom(self):
        current_w = self.root.winfo_reqwidth()
        current_h = self.root.winfo_reqheight()
        parent_w = self.parent_pos[2]
        parent_h = self.parent_pos[3]
        x = self.parent_pos[0] + (current_w - parent_w) / 2
        y = self.parent_pos[1] + (parent_h - current_h) / 2
        return '%dx%d+%d+%d' % (parent_w, parent_h, x, y)

    def _set_geom(self):
        geom = self._calculate_geom()
        # self.root.geometry(geom)
        self.root.geometry('+%s' % geom.split('+', 1)[1])

    def _configure_widgets(self):
        self.HelpLabel = elements.HelpLabel(
            self.root,
            row=1,
            col=2,
            link=HELP_LINK,
        )

        # Server URI
        self.server_label = tk.Label(self.root, text=_('Server URI'))
        self.layout_widget(self.server_label, 'server_label')

        server_entry_var = tk.StringVar()
        server_entry_var.trace('w',
                               lambda *args: character_limit(server_entry_var))

        self.server_entry = elements.CustomEntry(
            self.root,
            width=MAX_TEXT_LENGTH,
            textvariable=server_entry_var,
        )
        self.layout_widget(self.server_entry, 'server_entry')
        self.server_entry.insert(0, self.config_settings.cloud)

        self.entry_widgets[ConfigSettings.CLOUD] = self.server_entry

        # Domain name
        self.domain_label = tk.Label(self.root, text=_('Domain name'))
        self.layout_widget(self.domain_label, 'domain_label')

        domain_entry_var = tk.StringVar()
        domain_entry_var.trace('w',
                               lambda *args: character_limit(domain_entry_var))

        self.domain_entry = elements.CustomEntry(
            self.root,
            width=MAX_TEXT_LENGTH,
            textvariable=domain_entry_var,
        )
        self.layout_widget(self.domain_entry, 'domain_name')
        domain_name = self.config_settings.domain_name
        self.domain_entry.insert(0, domain_name)
        self.entry_widgets[ConfigSettings.DOMAIN_NAME] = self.domain_entry

        # Language
        self.language_label = tk.Label(self.root, text=_('Language'))
        self.layout_widget(self.language_label, 'language_label')

        self.language_entry = ttk.Combobox(
            self.root, values=Localization.locales_translate.values())

        self.layout_widget(self.language_entry, 'language_entry')

        option = Localization.locales_translate.get(
            self.config_settings.language)

        self.language_entry.set(option)
        self.entry_widgets[ConfigSettings.LANGUAGE] = self.language_entry

        # Buttons
        self.cancel_button = tk.Button(self.root,
                                       text=_('Cancel'),
                                       command=self.on_cancel_clicked)
        self.layout_widget(self.cancel_button, 'cancel_button')

        self.save_button = tk.Button(self.root,
                                     text=_('Save'),
                                     command=self.on_save_clicked)
        self.layout_widget(self.save_button, 'save_button')

    @staticmethod
    def layout_widget(widget, name):
        widgets_layout = {
            'server_label': WidgetLayout(row=2, column=1, sticky=tk.W),
            'server_entry': WidgetLayout(row=2, column=2, sticky=tk.W),
            'domain_label': WidgetLayout(row=3, column=1, sticky=tk.W),
            'domain_name': WidgetLayout(row=3, column=2, sticky=tk.W),
            'language_label': WidgetLayout(row=4, column=1, sticky=tk.W),
            'language_entry': WidgetLayout(row=4, column=2, sticky=tk.W),
            'save_button': WidgetLayout(row=5, column=2, sticky=tk.E),
            'cancel_button': WidgetLayout(row=5, column=1, sticky=tk.W),
        }
        widget_layout = widgets_layout.get(name) or WidgetLayout()
        widget.grid(
            row=widget_layout.row,
            column=widget_layout.column,
            sticky=widget_layout.sticky,
            padx=widget_layout.padx,
            pady=widget_layout.pady
        )

    def on_cancel_clicked(self):
        self.root.destroy()

    @catch_exceptions(fall=False)
    def on_save_clicked(self):
        # update params for saving
        new_language = None
        old_language = self.config_settings.language
        for setting_param in self.entry_widgets:
            if setting_param in ConfigSettings.ordered_file_variables:
                widget = self.entry_widgets.get(setting_param)
                value = widget.get()

                if isinstance(value, unicode):
                    value = value.encode('utf-8')

                if setting_param == ConfigSettings.LANGUAGE:
                    value = get_key_by_value(Localization.locales_translate,
                                             value.decode('utf-8'))
                setattr(self.config_settings, setting_param, value)
                new_language = self.config_settings.language

        self.config_settings.save()

        if new_language != old_language:
            msg = _('TIONIX.VDIclient should be restarted to '
                    'change localization.')
            messagebox.showinfo(message=msg)

        self.root.destroy()
        self.root = None
