# coding: utf-8

import binascii
import Queue
import logging
import os
import subprocess
import threading

from sys import stdout, platform, getfilesystemencoding

import requests

from tionix_vdi_client import MODULE_VERSION
from tionix_vdi_client import exceptions
from tionix_vdi_client.helpers import get_os_version
from tionix_vdi_client.i18n import ugettext as _
from tionix_vdi_client.splitters import get_valid_url
from tionix_vdi_client.settings import APP_DIR_PATH


logger = logging.getLogger(__name__)

CONNECTION_TYPE = 'client'
FAILED_GET_VM_IP_ERR_MSG = _(u"Couldn't get instance address.")


class VDIConnect(object):
    def __init__(self):
        super(VDIConnect, self).__init__()
        self.pipe = []

    def exec_all(self, *args, **kwargs):
        queue = Queue.Queue()
        kwargs['queue'] = queue

        threads_pool = []
        for target in self.pipe:
            thread = threading.Thread(
                target=target, args=args, kwargs=kwargs)
            thread.start()
            threads_pool.append(thread)

        for thread in threads_pool:
            thread.join()

        if not queue.empty():
            err_msg = queue.get()
            raise exceptions.ApplicationLogicException(err_msg)

    def add_exec(self, executer):
        assert callable(executer)
        self.pipe.append(executer)


class BaseExec(object):
    cmd_pipe = []
    bad_cmd_str = _("Invalid parameter: '{connect}'")
    connect_keys = (
        'ip', 'user', 'password', 'project', 'domain', 'rdp_config')

    def __init__(self, **params):
        super(BaseExec, self).__init__()
        self.params = {key: params.get(key, '') for key in self.connect_keys}

    def __call__(self, *args, **kwargs):
        for command in self.cmd_pipe:
            command(*args, **kwargs)


class StdoutExec(BaseExec):
    def __init__(self, **params):
        super(StdoutExec, self).__init__(**params)
        self.cmd_pipe = [self.echo_to_stdout]

    def echo_to_stdout(self, *args, **kwargs):
        lines = (
            u"username={user}".format(**self.params),
            u"password={password}".format(**self.params),
            u"ip={ip}\n".format(**self.params),
        )
        stdout.write('\n'.join(lines))


class VariousStrExec(BaseExec):
    def __init__(self, cmd_str, **params):
        super(VariousStrExec, self).__init__(**params)
        self.cmd_str = cmd_str

    def __call__(self, *args, **kwargs):
        self.queue = kwargs.get('queue')  # pylint:disable=W0201
        keys_to_escape = ['user', 'password', 'domain']
        escaped_params = self.escape_parameters(self.params, keys_to_escape)
        try:
            cmd = self.cmd_str.format(**escaped_params)
        except KeyError:
            msg = "Bad command line format %s" % self.cmd_str
            logger.critical(msg)
            self.queue.put(msg)
            raise exceptions.ApplicationLogicException(self.bad_cmd_str)

        self.run_proc(cmd)

    @staticmethod
    def escape_parameters(parameters, keys_to_escape):
        """
        Escapes specified values in parameters, depending on the OS.
        :param parameters: dict
        :param keys_to_escape: tuple/list
        :return: shallow copy of parameters with modified keys
        """

        def escape_linux(value):
            return "'%s'" % value.replace("'", "'\"'\"'")

        def escape_windows(value):
            return '"%s"' % value

        if platform == "win32":
            escaped = escape_windows
        else:
            escaped = escape_linux

        esc_params = parameters.copy()
        for key in keys_to_escape:
            esc_params[key] = escaped(esc_params[key])

        return esc_params

    def run_proc(self, command):
        logger.debug(u'Running command: {0}'.format(command))
        process = subprocess.Popen(
            command.encode(getfilesystemencoding()),
            shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        _out, err = process.communicate()
        if err:
            msg = _(u"Couldn't connect to "
                    u"instance: '{0}'").format(err.decode(stdout.encoding))
            self.queue.put(msg)
        return process


def get_response_json(response):
    """Return json dict from response"""

    if response.status_code == 500:
        logger.error(u"Error {0}. Url: {1}. Content: {2}".format(
            response.status_code, response.url, response.content))
        raise exceptions.VDIServerResponseError(_(u'Server error occurred.'))

    elif response.status_code == 409:
        logger.error(u"Error {0}. Content: {1}".format(
            response.status_code, unicode(response.content, 'utf-8')))
        msg = _(u"Instance is not ready. Please try to connect later.")
        raise exceptions.VDIServerResponseError(msg)

    elif response.status_code == 401:
        msg = _(u"Invalid login/password pair")
        raise exceptions.VDIServerBadCredentialsError(msg)

    elif response.status_code >= 400:
        logger.error(u"Error {0}. Content: {1}".format(response.status_code,
                                                       response.content))
        raise exceptions.VDIServerBadConnectionParamsError(
            FAILED_GET_VM_IP_ERR_MSG)

    try:
        json_body = response.json()
    except ValueError:
        logger.error(u"Response doesn't have json. Request url: {0}".format(
            response.url))
        raise exceptions.VDIServerResponseError(FAILED_GET_VM_IP_ERR_MSG)

    return json_body


def request_endpoint(srv):
    response = requests.get(srv)
    json_body = get_response_json(response)
    endpoint = json_body.get('tnx-vm-url')
    if not endpoint:
        logger.error(u"Couldn't get endpoint to retrieve instance address.")
        raise exceptions.VDIServerResponseError(FAILED_GET_VM_IP_ERR_MSG)
    return endpoint


def get_vm_ip(srv, user, password, project, domain):
    """
        Connects to server and retrieves instance IP address
        :returns: ip address
    """
    endpoint = request_endpoint(srv)

    client = u'TIONIX.VDIclient {0}'.format(MODULE_VERSION)
    payload = {
        "auth_user": user,
        "auth_password": password,
        "auth_tenant": project,
        "project_domain_name": domain,
        "user_domain_name": domain,
        "conn_type": CONNECTION_TYPE,
        "os": get_os_version(),
        "client": client,
    }

    response = requests.get(endpoint, params=payload)
    json_body = get_response_json(response)
    vm_ip = json_body.get('ip')
    if not vm_ip:
        logger.error(u"Couldn't get ip from json")
        raise exceptions.VDIServerResponseError(FAILED_GET_VM_IP_ERR_MSG)
    logger.debug(u"Got ip: {0}".format(vm_ip))
    return vm_ip


def get_connection_params(conn_credentials, config_settings):

    def check_required_values(parameters):
        """
        Checks required keys for empty values.
        :param parameters: dict or Settings()
        :raises: ApplicationLogicException
        """
        required_keys = getattr(
            parameters, 'required_attributes', parameters.keys())
        for key, value in parameters.items():
            if key in required_keys and not value:
                raise exceptions.MissingParameterException(
                    (_(u"Parameter \"{0}\" required")).format(key)
                )

    check_required_values(conn_credentials)
    check_required_values(config_settings)

    user = conn_credentials['user']
    password = conn_credentials['password']
    project = conn_credentials['project']
    domain_name = conn_credentials['domain_name']

    srv = get_valid_url(config_settings.cloud)

    logger.debug("Trying to connect to {0} as user: {1}".format(srv, user))

    connection_params = {
        'srv': srv,
        'user': user,
        'password': password,
        'project': project,
        'domain': domain_name,
    }
    return connection_params


def genetate_win_rdp_config(host, domain, user, password):
    rdp_config_template = u"""
full address:s:{host}
UserName:s:{domain}\{user}
password 51:b:{password_encrypted}
"""

    def get_encrypted_rdp_password(raw_password):
        from tionix_vdi_client.winapi import crypt_protect_data

        pass_encrypted = crypt_protect_data(raw_password)
        return binascii.hexlify(pass_encrypted).upper()

    password_encrypted = get_encrypted_rdp_password(password)
    content = rdp_config_template.format(
        host=host,
        domain=domain,
        user=user,
        password_encrypted=password_encrypted
    )

    rdp_config_path = os.path.join(APP_DIR_PATH, 'connect.rdp')
    with open(rdp_config_path, 'w') as fh:
        fh.write(content.encode('utf-8'))

    return rdp_config_path.decode(getfilesystemencoding())


def vdi_connect(conn_credentials, config_settings):
    """For user credentials connect to remote instance"""
    conn_params = get_connection_params(conn_credentials, config_settings)

    try:
        dest_ip = get_vm_ip(**conn_params)
    except requests.exceptions.ConnectionError:
        raise exceptions.ApplicationConnectionError(
            _(u"Couldn't connect to server: {}").format(conn_params['srv']))

    except requests.exceptions.InvalidURL:
        raise exceptions.ApplicationConnectionError(
            _(u"Invalid address: {}").format(conn_params['srv']))

    conn_cmd_params = {}
    for key in ("user", "password", "project", "domain"):
        conn_cmd_params[key] = conn_params[key].decode('utf-8')
    conn_cmd_params['ip'] = dest_ip

    logger.debug("Running connections to {0} as user: "
                 "{1}".format(conn_params['srv'], conn_params['user']))

    conn_cmd_template = unicode(config_settings.connect, 'utf-8')

    if platform == "win32":
        # on Windows platform some special characters in password can't be
        # escaped correctly, so we generate RDB-file with connection
        # parameters to pass to rdp client
        conn_cmd_params['rdp_config'] = genetate_win_rdp_config(
            host=conn_cmd_params['ip'],
            domain=conn_cmd_params['domain'],
            user=conn_cmd_params['user'],
            password=conn_cmd_params['password']
        )

    vdi = VDIConnect()
    vdi.add_exec(StdoutExec(**conn_cmd_params))
    vdi.add_exec(VariousStrExec(conn_cmd_template, **conn_cmd_params))
    vdi.exec_all()
